@@ -9,12 +9,12 @@ import logging
 from logging import NullHandler
 
 from .nem_reader import NEMFile, read_nem_file
+from .output_db import extend_sqlite, output_as_sqlite, output_folder_as_sqlite
 from .outputs import (
     nmis_in_file,
     output_as_csv,
     output_as_daily_csv,
     output_as_data_frames,
-    output_as_sqlite,
 )
 from .version import __version__
 
@@ -27,6 +27,8 @@ __all__ = [
     "output_as_daily_csv",
     "output_as_data_frames",
     "output_as_sqlite",
+    "output_folder_as_sqlite",
+    "extend_sqlite",
 ]
 
 # Set default logging handler to avoid "No handler found" warnings.
