@@ -24,16 +24,17 @@ from .split_days import make_set_interval, split_multiday_reads
 
 log = logging.getLogger(__name__)
 
+minutes_per_day = 24 * 60
+
 
 class NEMFile:
     """An NEM file object"""
 
     def __init__(self, file_path: str, strict: bool = False) -> None:
-
         self.file_path = file_path
         self.strict = strict
-        self.nmis: set = set()
-        self.nmi_channels: dict = {}
+        self._nmis: set = set()
+        self._nmi_channels: dict = {}
 
     def __repr__(self):
         return f"<NEMFile {self.file_path}>"
@@ -46,6 +47,22 @@ class NEMFile:
             return True
         return False
 
+    @property
+    def nmis(self) -> set:
+        """NMIs in file"""
+        if self._nmis:
+            return self._nmis
+        self.nem_data()  # Need to process file first
+        return self._nmis
+
+    @property
+    def nmi_channels(self) -> dict:
+        """NMI channels in file"""
+        if self._nmi_channels:
+            return self._nmi_channels
+        self.nem_data()  # Need to process file first
+        return self._nmi_channels
+
     def parse_nem_file(self, nem_file, file_name="") -> NEMReadings:
         """Parse NEM file and return meter readings named tuple"""
         reader = csv.reader(nem_file, delimiter=",")
@@ -94,32 +111,32 @@ class NEMFile:
                     # Zip file is open in binary mode
                     # So decode then convert back to list
                     nmi_file = csv_text.read().decode("utf-8").splitlines()
-                    reads = self.parse_nem_file(nmi_file, file_name=csv_file)
-                    for nmi in reads.transactions.keys():
-                        self.nmis.add(nmi)
-                        suffixes = list(reads.transactions[nmi].keys())
-                        self.nmi_channels[nmi] = suffixes
-                    return NEMData(
-                        header=self.header,
-                        readings=reads.readings,
-                        transactions=reads.transactions,
-                    )
+                reads = self.parse_nem_file(nmi_file, file_name=csv_file)
+                for nmi in reads.transactions.keys():
+                    self._nmis.add(nmi)
+                    suffixes = list(reads.transactions[nmi].keys())
+                    self._nmi_channels[nmi] = suffixes
+                return NEMData(
+                    header=self.header,
+                    readings=reads.readings,
+                    transactions=reads.transactions,
+                )
 
         with open(self.file_path) as nmi_file:
             reads = self.parse_nem_file(nmi_file)
-            for nmi in reads.transactions.keys():
-                self.nmis.add(nmi)
-                suffixes = list(reads.transactions[nmi].keys())
-                self.nmi_channels[nmi] = suffixes
-            return NEMData(
-                header=self.header,
-                readings=reads.readings,
-                transactions=reads.transactions,
-            )
+        for nmi in reads.transactions.keys():
+            self._nmis.add(nmi)
+            suffixes = list(reads.transactions[nmi].keys())
+            self._nmi_channels[nmi] = suffixes
+        return NEMData(
+            header=self.header,
+            readings=reads.readings,
+            transactions=reads.transactions,
+        )
 
     def get_data_frame(
         self, split_days: bool = False, set_interval: Optional[int] = None
-    ) -> pd.DataFrame:
+    ) -> Optional[pd.DataFrame]:
         """Return NEMData as a DataFrame"""
         nd = self.nem_data()
         frames = []
@@ -137,14 +154,16 @@ class NEMFile:
                     "nmi": [nmi for _ in range(len(reads))],
                     "suffix": [suffix for _ in range(len(reads))],
                     "serno": [x.meter_serial_number for x in reads],
-                    "t_start": [x.t_start for x in reads],
-                    "t_end": [x.t_end for x in reads],
+                    "t_start": pd.to_datetime([x.t_start for x in reads]),
+                    "t_end": pd.to_datetime([x.t_end for x in reads]),
                     "value": [x.read_value for x in reads],
                     "quality": [x.quality_method for x in reads],
                     "evt_code": [x.event_code for x in reads],
                     "evt_desc": [x.event_desc for x in reads],
                 }
                 frames.append(pd.DataFrame(data))
+        if not frames:
+            return None
         return pd.concat(frames)
 
     def get_pivot_data_frame(
@@ -152,18 +171,53 @@ class NEMFile:
         split_days: bool = False,
         set_interval: Optional[int] = None,
         include_serno: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Optional[pd.DataFrame]:
         """Return NEMData as a DataFrame with suffix columns"""
         df = self.get_data_frame(split_days, set_interval)
-        pivot_index = ["nmi", "t_start", "t_end", "quality", "evt_code", "evt_desc"]
+        if df is None:
+            return df
+        index_cols = [
+            "nmi",
+            "suffix",
+            "t_start",
+            "t_end",
+        ]
         if include_serno:
-            pivot_index.append("serno")
-        df_pivoted = df.pivot(
-            index=pivot_index,
-            columns="suffix",
-            values="value",
-        ).reset_index()
-        return df_pivoted
+            index_cols.append("serno")
+        else:
+            del df["serno"]
+
+        df.set_index(index_cols, inplace=True)
+        df = df.unstack("suffix")
+        df = df.reset_index()
+        df.columns = df.columns.map("_".join).str.strip("_")
+        quality = 0
+        evt_code = 0
+        evt_desc = 0
+        new_names = {}
+        for col in df.columns:
+            if "value_" in col:
+                new_names[col] = col[6:]
+            if "quality" in col:
+                if not quality:
+                    new_names[col] = "quality"
+                else:
+                    del df[col]
+                quality += 1
+            if "evt_code" in col:
+                if not evt_code:
+                    new_names[col] = "evt_code"
+                else:
+                    del df[col]
+                evt_code += 1
+            if "evt_desc" in col:
+                if not evt_desc:
+                    new_names[col] = "evt_desc"
+                else:
+                    del df[col]
+                evt_desc += 1
+        df.rename(columns=new_names, inplace=True)
+        return df
 
     def get_per_nmi_dfs(
         self,
@@ -179,9 +233,9 @@ class NEMFile:
             yield nmi, nmi_df
 
 
-def flatten_list(l: List[list]) -> list:
+def flatten_list(items: List[list]) -> list:
     """takes a list of lists, l and returns a flat list"""
-    return [v for inner_l in l for v in inner_l]
+    return [v for inner_l in items for v in inner_l]
 
 
 def read_nem_file(file_path: str, ignore_missing_header=False) -> NEMData:
@@ -198,7 +252,11 @@ def read_nem_file(file_path: str, ignore_missing_header=False) -> NEMData:
 
 
 def parse_100_row(row: List[Any], file_name: str) -> HeaderRecord:
-    """Parse header record (100)"""
+    """Parse header record (100)
+
+    RecordIndicator,VersionHeader,DateTime,FromParticipant,ToParticipant
+    Example: 100,NEM12,200301011534,MDP1,Retailer1
+    """
     return HeaderRecord(
         row[1], parse_datetime(row[2]), row[3], row[4], file_name, False
     )
@@ -212,81 +270,94 @@ def parse_nem12_rows(nem_list: Iterable, file_name=None) -> NEMReadings:
     trans: Dict[str, Dict[str, list]] = {}
     nmi_d = None  # current NMI details block that readings apply to
 
-    observed_900_record = False
+    observed_900_records = []
 
-    for row in nem_list:
-
-        if not row:
-            log.debug("Skipping empty row.")
-            continue
-
-        record_indicator = int(row[0])
-
-        if record_indicator == 900:
-            # Powercor NEM12 files can concatenate multiple files together
-            # try to keep parsing anyway.
-            if observed_900_record:
-                log.warning("Found multiple end of data (900) rows. ")
-
-            observed_900_record = True
-            pass
-
-        elif record_indicator == 200:
-            try:
-                nmi_details = parse_200_row(row)
-            except ValueError:
-                log.error("Error passing 200 row:")
-                log.error(row)
-                raise
-            nmi_d = nmi_details
-
-            if nmi_d.nmi not in readings:
-                readings[nmi_d.nmi] = {}
-            if nmi_d.nmi_suffix not in readings[nmi_d.nmi]:
-                readings[nmi_d.nmi][nmi_d.nmi_suffix] = []
-            if nmi_d.nmi not in trans:
-                trans[nmi_d.nmi] = {}
-            if nmi_d.nmi_suffix not in trans[nmi_d.nmi]:
-                trans[nmi_d.nmi][nmi_d.nmi_suffix] = []
-
-        elif record_indicator == 300:
-            num_intervals = int(24 * 60 / nmi_d.interval_length)
-            assert len(row) > 1, f"Invalid 300 Row in {file_name}"
-            if len(row) < num_intervals + 2:
-                record_date = row[1]
-                msg = "Skipping 300 record for %s %s %s. "
-                msg += "It does not have the expected %s intervals"
-                log.error(msg, nmi_d.nmi, nmi_d.nmi_suffix, record_date, num_intervals)
+    for row_num, row in enumerate(nem_list, start=1):
+        try:
+            if not row:
+                log.debug(f"Skipping empty row at line {row_num}.")
                 continue
-            interval_record = parse_300_row(
-                row, nmi_d.interval_length, nmi_d.uom, nmi_d.meter_serial_number
-            )
-            # don't flatten the list of interval readings at this stage,
-            # as they may need to be adjusted by a 400 row
-            readings[nmi_d.nmi][nmi_d.nmi_suffix].append(
-                interval_record.interval_values
-            )
 
-        elif record_indicator == 400:
-            event_record = parse_400_row(row)
-            readings[nmi_d.nmi][nmi_d.nmi_suffix][-1] = update_reading_events(
-                readings[nmi_d.nmi][nmi_d.nmi_suffix][-1], event_record
-            )
+            record_indicator = int(row[0])
 
-        elif record_indicator == 500:
-            b2b_details = parse_500_row(row)
-            trans[nmi_d.nmi][nmi_d.nmi_suffix].append(b2b_details)
+            if record_indicator == 900:
+                # Powercor NEM12 files can concatenate multiple files together
+                # try to keep parsing anyway.
+                if observed_900_records:
+                    log.warning(
+                        f"Found multiple end of data (900) rows on lines {observed_900_records}"
+                    )
 
-        else:
-            log.warning(
-                "Record indicator %s not supported and was skipped", record_indicator
-            )
+                observed_900_records.append(row_num)
+                pass
+
+            elif record_indicator == 200:
+                try:
+                    nmi_details = parse_200_row(row)
+                except ValueError:
+                    log.error(f"Error passing 200 row at line {row_num}:")
+                    log.error(row)
+                    raise
+                nmi_d = nmi_details
+
+                if nmi_d.nmi not in readings:
+                    readings[nmi_d.nmi] = {}
+                if nmi_d.nmi_suffix not in readings[nmi_d.nmi]:
+                    readings[nmi_d.nmi][nmi_d.nmi_suffix] = []
+                if nmi_d.nmi not in trans:
+                    trans[nmi_d.nmi] = {}
+                if nmi_d.nmi_suffix not in trans[nmi_d.nmi]:
+                    trans[nmi_d.nmi][nmi_d.nmi_suffix] = []
+
+            elif record_indicator == 300:
+                num_intervals = int(minutes_per_day / nmi_d.interval_length)
+                assert len(row) > 1, f"Invalid 300 Row in {file_name} on line {row_num}"
+                if len(row) < num_intervals + 2:
+                    record_date = row[1]
+                    msg = "Skipping 300 record for %s %s %s on row %d. "
+                    msg += "It does not have the expected %s intervals"
+                    log.error(
+                        msg,
+                        nmi_d.nmi,
+                        nmi_d.nmi_suffix,
+                        record_date,
+                        row_num,
+                        num_intervals,
+                    )
+                    continue
+                interval_record = parse_300_row(
+                    row, nmi_d.interval_length, nmi_d.uom, nmi_d.meter_serial_number
+                )
+                # don't flatten the list of interval readings at this stage,
+                # as they may need to be adjusted by a 400 row
+                readings[nmi_d.nmi][nmi_d.nmi_suffix].append(
+                    interval_record.interval_values
+                )
+
+            elif record_indicator == 400:
+                event_record = parse_400_row(row, nmi_d.interval_length)
+                readings[nmi_d.nmi][nmi_d.nmi_suffix][-1] = update_reading_events(
+                    readings[nmi_d.nmi][nmi_d.nmi_suffix][-1], event_record
+                )
+
+            elif record_indicator == 500:
+                b2b_details = parse_500_row(row)
+                trans[nmi_d.nmi][nmi_d.nmi_suffix].append(b2b_details)
+
+            else:
+                log.warning(
+                    "Record indicator %s on line %d not supported and was skipped",
+                    record_indicator,
+                    row_num,
+                )
+        except (KeyError, ValueError, AssertionError, IndexError, TypeError) as e:
+            raise ValueError(f"Unable to parse line {row_num}") from e
 
     for nmi in readings:
         for suffix in readings[nmi]:
             readings[nmi][suffix] = flatten_list(readings[nmi][suffix])
 
-    if not observed_900_record:
+    if not observed_900_records:
         log.warning("Missing end of data (900) row.")
 
     return NEMReadings(readings=readings, transactions=trans)
@@ -364,7 +435,11 @@ def calculate_manual_reading(basic_data: BasicMeterData) -> Reading:
 
 
 def parse_200_row(row: list) -> NmiDetails:
-    """Parse NMI data details record (200)"""
+    """Parse NMI data details record (200)
+
+    RecordIndicator,NMI,NMIConfiguration,RegisterID,NMISuffix,MDMDataStreamIdentifier,MeterSerialNumber,UOM,IntervalLength,NextScheduledReadDate
+    Example: 200,VABD000163,E1Q1,1,E1,N1,METSER123,kWh,30,20040120
+    """
 
     next_read = None  # Next scheduled read is an optional field
     if len(row) > 9:
@@ -376,7 +451,15 @@ def parse_200_row(row: list) -> NmiDetails:
 
 
 def parse_250_row(row: list) -> BasicMeterData:
-    """Parse basic meter data record (250)"""
+    """Parse basic meter data record (250)
+
+        RecordIndicator,NMI,NMIConfiguration,RegisterID,NMISuffix,MDMDataStreamIdentifier,MeterSeri
+    alNumber,DirectionIndicator,PreviousRegisterRead,PreviousRegisterReadDateTime,PreviousQuali
+    tyMethod,PreviousReasonCode,PreviousReasonDescription,CurrentRegisterRead,CurrentRegister
+    ReadDateTime,CurrentQualityMethod,CurrentReasonCode,CurrentReasonDescription,Quantity,U
+    OM,NextScheduledReadDate,UpdateDateTime,MSATSLoadDateTime
+        Example: 250,1234567890,1141,01,11,11,METSER66,E,000021.2,20031001103230,A,,,000534.5,20040201100030,E64,77,,343.5,kWh,20040509, 20040202125010,20040203000130
+    """
     return BasicMeterData(
         row[1],
         row[2],
@@ -411,11 +494,27 @@ def nth(iterable, n, default=None):
 def parse_300_row(
     row: list, interval: int, uom: str, meter_serial_number: str
 ) -> IntervalRecord:
-    """Interval data record (300)"""
+    """Interval data record (300)
+
+    RecordIndicator,IntervalDate,IntervalValue1 . . . IntervalValueN,
+    QualityMethod,ReasonCode,ReasonDescription,UpdateDateTime,MSATSLoadDateTime
+    Example: 300,20030501,50.1, . . . ,21.5,V,,,20030101153445,20030102023012
+    """
+
+    # count of fields except IntervalValue1 . . . IntervalValueN
+    # excluding MSATSLoadDateTime which is only required if present
+    num_required_non_reading_fields = 6
+
+    num_intervals = int(minutes_per_day / interval)
 
-    num_intervals = int(24 * 60 / interval)
     interval_date = parse_datetime(row[1])
     last_interval = 2 + num_intervals
+    if len(row) < num_intervals + num_required_non_reading_fields:
+        num_rows = len(row) - num_required_non_reading_fields
+        raise ValueError(
+            f"Unexpected number of values in 300 row: " +
+            f"{num_rows} readings for {interval}min intervals"
+        )
     quality_method = row[last_interval]
 
     # Optional fields
@@ -486,8 +585,29 @@ def parse_reading(val: str) -> Optional[float]:
         return None  # Not a valid reading
 
 
-def parse_400_row(row: list) -> tuple:
-    """Interval event record (400)"""
+def parse_400_row(row: list, interval_length: int) -> tuple:
+    """Interval event record (400)
+
+    RecordIndicator,StartInterval,EndInterval,QualityMethod,ReasonCode,ReasonDescription
+    Example: 400,1,28,S14,32
+
+    Note that intervals are indexed from 1 not 0.
+    """
+    num_intervals = int(minutes_per_day / interval_length)
+    start_interval = int(row[1])
+    end_interval = int(row[2])
+    if end_interval < start_interval:
+        raise ValueError(
+            f"End interval {end_interval} is earlier than start interval {start_interval} in 400 row."
+        )
+    if not (0 < start_interval <= num_intervals):
+        raise ValueError(
+            f"Invalid start interval {start_interval} in 400 row. Expecting {num_intervals} intervals."
+        )
+    if not (0 < end_interval <= num_intervals):
+        raise ValueError(
+            f"Invalid end interval {end_interval} in 400 row. Expecting {num_intervals} intervals."
+        )
 
     return EventRecord(int(row[1]), int(row[2]), row[3], row[4], row[5])
 
@@ -514,12 +634,20 @@ def update_reading_events(readings, event_record):
 
 
 def parse_500_row(row: list) -> tuple:
-    """Parse B2B details record"""
+    """Parse B2B details record
+
+    RecordIndicator,TransCode,RetServiceOrder,ReadDateTime,IndexRead
+    Example: 500,S,RETNSRVCEORD1,20031220154500,001123.5
+    """
     return B2BDetails12(row[1], row[2], row[3], row[4])
 
 
 def parse_550_row(row: list) -> tuple:
-    """Parse B2B details record"""
+    """Parse B2B details record
+
+    RecordIndicator,PreviousTransCode,PreviousRetServiceOrder,CurrentTransCode,CurrentRetServiceOrder
+    Example: 550,N,,A,
+    """
     return B2BDetails13(row[1], row[2], row[3], row[4])
 
 
