@@ -1,6 +1,3 @@
-""" Test Suite
-"""
-
 import pytest
 
 from nemreader import NEMFile
