@@ -1,6 +1,3 @@
-""" Test Suite
-"""
-
 import pandas as pd
 
 from nemreader.outputs import output_as_data_frames
